-- stream.ma -- MiniAgda stream library

cofun Stream : ++(A : Set) -> -(i : Size) -> Set
{ Stream A i = A & ([i' < i] -> Stream A i')
}

let inst [A : Set] [i : Size] (f : [j < $i] -> Stream A j) : Stream A i
  = \ j -> (f $j) j

fun head : [A : Set] [i : Size] (s : Stream A i) -> A
{ head A i (a , as) = a
}
fun tail : [A : Set] [i : Size] (s : Stream A $i) -> Stream A i
{ tail A i (a , as) = as i
}

fun sTake : [A : Set] [i : Size] (n : Nat i)(s : Stream A i) -> List A i
{ sTake A i zero          s        = nil
; sTake A i (suc (i', n)) (a , as) = cons a (i', sTake A i' n (as i'))
}

let sDrop [A : Set] (n : Nat #) (s : Stream A #) : Stream A #
  = iterNat (Stream A #) (tail A #) s # n

fun zipWith2 : [A, B, C : Set] (f : A -> B -> C) -> 
  [i : Size] -> |i| -> Stream A i -> Stream B i -> Stream C i
{ zipWith2 A B C f i (a , as) (b , bs) = f a b , \ j -> 
    zipWith2 A B C f j (as j) (bs j) 
}

-- Bisimilarity

cofun StreamEq : [A : Set] (as, bs : Stream A #) -(i : Size) -> Set
{ StreamEq A as bs i = Id A (head A # as) (head A # bs) &
   ([i' < i] -> StreamEq A (tail A # as) (tail A # bs) i') 
}

-- Defining the n-ary zipwith

data MAYBE [a : Size] ++(A : Set a) : Set a
{ Nothing : MAYBE a A
; Just    : A -> MAYBE a A 
}

cofun LIST : [a : Size] ++(A : Set a) +(i : Size) -> Set a 
{ LIST a A i = MAYBE a ([i' < i] & A & LIST a A i')
}

fun Map : [a : Size] [A, B : Set a] (f : A -> B) 
  [i : Size] -> LIST a A i -> LIST a B i
{ Map a' A B f i Nothing = Nothing
; Map a' A B f i (Just (i' , a , as)) = Just (i' , f a , Map a' A B f i' as)
}

fun Fun : [i : Size] -> LIST $0 Set i -> Set -> Set 
{ Fun i Nothing             C = C
; Fun i (Just (i' , A , L)) C = A -> Fun i' L C
}

let Sets +(i : Size) = LIST $0 Set i
let MapSets = Map $0 Set Set

--  zipWith' []    c k          = c , k
--  zipWith' (A:L) f k (a , as) = zipWith' L (f a) (k as) 

fun zipWith' : [i : Size] |i| (L : Sets i) [C : Set] (f : Fun i L C)
  [j : Size] 
  (k : [j' < j] -> 
       Fun i (MapSets (\ A -> Stream A j') i L) (Stream C j')) ->
  Fun i (MapSets (\ A -> Stream A j) i L) (Stream C j)
{ zipWith' i Nothing C c j cs = c , cs
; zipWith' i (Just (i' , A , L)) C f j k (a , as) =
   zipWith' i' L C (f a) j (\ j' -> k j' (as j'))
}

--  zipWith L f = zipWith' L f (zipWith L f)

fun zipWith : [i : Size] (L : Sets i) [C : Set] (f : Fun i L C)
  [j : Size] -> |j| -> Fun i (MapSets (\ A -> Stream A j) i L) (Stream C j) 
{ zipWith i L C f j = zipWith' i L C f j (zipWith i L C f) }

let repeat (A : Set)(a : A)[i : Size] : Stream A i
  = zipWith 0 Nothing A a i

let sMap (A, B : Set)(f : A -> B)[i : Size](s : Stream A i) : Stream B i
  = zipWith 1 (Just (0, A, Nothing)) B f i s

let sZipWith (A, B, C : Set)
  = zipWith 2 (Just (1, A, Just (0, B, Nothing))) C

fun iterate : [A : Set](f : A -> A)(a : A)[i : Size] |i| -> Stream A i
{ iterate A f a i = a , \ i' -> iterate A f (f a) i'
}

let upFrom : (n : Nat #)[i : Size] -> Stream (Nat #) i
  = iterate (Nat #) (succ #)
 
-- nested coinduction?

fun bla : (A : Set)(f : A -> A)(a : A)[i,j : Size] |i,j| -> 
  Stream (Stream A j) i
{ 
{-
bla A a i j = mapStream ([j' < j] -> Stream A j') (Stream A j) 
   (\ as -> a , as) i (bla A a i j') -- where to get j' from?
; 
bla A f a i j = iterate A f a j , \ i' -> 
  sMap (Stream A #) (Stream A #) (tail A #) i' (bla A f a i' #)
-}
bla A f a i j = iterate A f a j , \ i' -> 
  sZipWith (Nat #) (Stream A #) (Stream A #) (sDrop A) 
    i' (upFrom (suc (0, zero)) i') (bla A f a i' #)
}

let squares = sMap (Stream (Nat #) #) (Nat #) (head (Nat #) #) 
                # (bla (Nat #) (succ #) zero # #) 

eval let sq4 = sTake (Nat #) # fourN squares  -- [0,1,3,6]

{-
sq4 evaluates to 
(true , ((false , unit) , (3 , 
(true , ((true , (# , (false , unit))) , (2 , 
(true , ((true , (# , (true , (# , (true , (# , (false , unit))))))) , (1 , 
(true , ((true , (# , (true , (# , (true , (# , (true , (# , (true , (# , (true , (# , (false , unit))))))))))))) , (0 , (false , unit)))))))))))))
-}